/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Tue Apr 10 2012
 */
#include "acgenomix.h"

#include <sys/select.h>

#include <tcl.h>

#include "genomix.h"


/* --- local data ---------------------------------------------------------- */

static fd_set watched;

static void	gnmx_event_proc(ClientData clientData, int mask);


/* --- gnmx_watch_init ----------------------------------------------------- */

void
gnmx_watch_init(void)
{
  FD_ZERO(&watched);
}


/* --- gnmx_watch ---------------------------------------------------------- */

void
gnmx_watch(struct gnmx_handle *h)
{
  if (FD_ISSET(h->eventfd, &watched)) return;

  FD_SET(h->eventfd, &watched);
  Tcl_CreateFileHandler(h->eventfd, TCL_READABLE,
                        gnmx_event_proc, (ClientData)h);
}


/* --- gnmx_unwatch -------------------------------------------------------- */

void
gnmx_unwatch(struct gnmx_handle *h)
{
  struct gnmx_handle *i;

  /* look for other instances using the same fd */
  for(i = gnmx_clients; i; i = i->next) {
    if (i == h) continue;
    if (i->eventfd == h->eventfd) return /* fd still in use */;
  }

  /* no one is using this fd anymore */
  Tcl_DeleteFileHandler(h->eventfd);
  FD_CLR(h->eventfd, &watched);
}


/* --- gnmx_event_proc ----------------------------------------------------- */

static void
gnmx_event_proc(ClientData clientData, int mask)
{
  int fd = ((struct gnmx_handle *)clientData)->eventfd;
  struct gnmx_handle *i;
  int s;

  for(i = gnmx_clients; i; i = i->next) {
    if (i->eventfd != fd) continue;

    Tcl_Preserve(i->interp);
    gnmx_atexit_push(i->dlh);
    s = i->info->doevents(i->c);
    gnmx_atexit_pop();
    if (s) Tcl_BackgroundError(i->interp);
    Tcl_Release(i->interp);
  }
}
